from recap.converters.protobuf import ProtobufConverter
from recap.types import (
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    StringType,
    StructType,
    UnionType,
)


def test_protobuf_converter():
    protobuf_schema = """
    syntax = "proto3";
    message Test {
        string name = 1;
        bool is_valid = 2;
        bytes data = 3;
        int32 id = 4;
        int64 large_id = 5;
        uint32 uid = 6;
        uint64 large_uid = 7;
        sint32 sid = 8;
        sint64 large_sid = 9;
        float score = 10;
        double large_score = 11;
        fixed32 fixed_id = 12;
        fixed64 large_fixed_id = 13;
        sfixed32 sfixed_id = 14;
        sfixed64 large_sfixed_id = 15;
    }
    """
    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert isinstance(recap_schema, StructType)
    assert recap_schema.alias == "_root.Test"
    fields = recap_schema.fields
    assert len(fields) == 15

    # Field 1: string name = 1;
    assert isinstance(fields[0], UnionType)
    assert fields[0].extra_attrs["name"] == "name"
    assert fields[0].extra_attrs["default"] == None
    assert isinstance(fields[0].types[1], StringType)
    assert fields[0].types[1].bytes_ == 2_147_483_648
    assert fields[0].types[1].variable == True

    # Field 2: bool is_valid = 2;
    assert isinstance(fields[1], UnionType)
    assert fields[1].extra_attrs["name"] == "is_valid"
    assert fields[1].extra_attrs["default"] == None
    assert isinstance(fields[1].types[1], BoolType)

    # Field 3: bytes data = 3;
    assert isinstance(fields[2], UnionType)
    assert fields[2].extra_attrs["name"] == "data"
    assert fields[2].extra_attrs["default"] == None
    assert isinstance(fields[2].types[1], BytesType)
    assert fields[2].types[1].bytes_ == 2_147_483_648
    assert fields[2].types[1].variable == True

    # Field 4: int32 id = 4;
    assert isinstance(fields[3], UnionType)
    assert fields[3].extra_attrs["name"] == "id"
    assert fields[3].extra_attrs["default"] == None
    assert isinstance(fields[3].types[1], IntType)
    assert fields[3].types[1].bits == 32
    assert fields[3].types[1].signed == True

    # Field 5: int64 large_id = 5;
    assert isinstance(fields[4], UnionType)
    assert fields[4].extra_attrs["name"] == "large_id"
    assert fields[4].extra_attrs["default"] == None
    assert isinstance(fields[4].types[1], IntType)
    assert fields[4].types[1].bits == 64
    assert fields[4].types[1].signed == True

    # Field 6: uint32 uid = 6;
    assert isinstance(fields[5], UnionType)
    assert fields[5].extra_attrs["name"] == "uid"
    assert fields[5].extra_attrs["default"] == None
    assert isinstance(fields[5].types[1], IntType)
    assert fields[5].types[1].bits == 32
    assert fields[5].types[1].signed == False

    # Field 7: uint64 large_uid = 7;
    assert isinstance(fields[6], UnionType)
    assert fields[6].extra_attrs["name"] == "large_uid"
    assert fields[6].extra_attrs["default"] == None
    assert isinstance(fields[6].types[1], IntType)
    assert fields[6].types[1].bits == 64
    assert fields[6].types[1].signed == False

    # Field 8: sint32 sid = 8;
    assert isinstance(fields[7], UnionType)
    assert fields[7].extra_attrs["name"] == "sid"
    assert fields[7].extra_attrs["default"] == None
    assert isinstance(fields[7].types[1], IntType)
    assert fields[7].types[1].bits == 32
    assert fields[7].types[1].signed == True

    # Field 9: sint64 large_sid = 9;
    assert isinstance(fields[8], UnionType)
    assert fields[8].extra_attrs["name"] == "large_sid"
    assert fields[8].extra_attrs["default"] == None
    assert isinstance(fields[8].types[1], IntType)
    assert fields[8].types[1].bits == 64
    assert fields[8].types[1].signed == True

    # Field 10: float score = 10;
    assert isinstance(fields[9], UnionType)
    assert fields[9].extra_attrs["name"] == "score"
    assert fields[9].extra_attrs["default"] == None
    assert isinstance(fields[9].types[1], FloatType)
    assert fields[9].types[1].bits == 32

    # Field 11: double large_score = 11;
    assert isinstance(fields[10], UnionType)
    assert fields[10].extra_attrs["name"] == "large_score"
    assert fields[10].extra_attrs["default"] == None
    assert isinstance(fields[10].types[1], FloatType)
    assert fields[10].types[1].bits == 64

    # Field 12: fixed32 fixed_id = 12;
    assert isinstance(fields[11], UnionType)
    assert fields[11].extra_attrs["name"] == "fixed_id"
    assert fields[11].extra_attrs["default"] == None
    assert isinstance(fields[11].types[1], IntType)
    assert fields[11].types[1].bits == 32
    assert fields[11].types[1].signed == False

    # Field 13: fixed64 large_fixed_id = 13;
    assert isinstance(fields[12], UnionType)
    assert fields[12].extra_attrs["name"] == "large_fixed_id"
    assert fields[12].extra_attrs["default"] == None
    assert isinstance(fields[12].types[1], IntType)
    assert fields[12].types[1].bits == 64
    assert fields[12].types[1].signed == False

    # Field 14: sfixed32 sfixed_id = 14;
    assert isinstance(fields[13], UnionType)
    assert fields[13].extra_attrs["name"] == "sfixed_id"
    assert fields[13].extra_attrs["default"] == None
    assert isinstance(fields[13].types[1], IntType)
    assert fields[13].types[1].bits == 32
    assert fields[13].types[1].signed == True

    # Field 15: sfixed64 large_sfixed_id = 15;
    assert isinstance(fields[14], UnionType)
    assert fields[14].extra_attrs["name"] == "large_sfixed_id"
    assert fields[14].extra_attrs["default"] == None
    assert isinstance(fields[14].types[1], IntType)
    assert fields[14].types[1].bits == 64
    assert fields[14].types[1].signed == True


def test_protobuf_converter_repeated():
    protobuf_schema = """
        syntax = "proto3";
        message Test {
            repeated int32 values = 1;
        }
    """

    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    ListType(IntType(bits=32, signed=True)),
                ],
                name="values",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_protobuf_converter_map():
    protobuf_schema = """
        syntax = "proto3";
        message Test {
            map<string, int32> value = 1;
        }
    """

    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert isinstance(recap_schema, StructType)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    MapType(
                        StringType(variable=True, bytes_=2_147_483_648),
                        IntType(bits=32, signed=True),
                    ),
                ],
                name="value",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_protobuf_converter_forward_reference():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            Inner outerValue = 1;
        }

        message Inner {
            int32 innerValue = 1;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)

    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="_root.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="_root.Outer",
    )


def test_protobuf_converter_map_forward_reference():
    protobuf_schema = """
        syntax = "proto3";
        message Test {
            map<string, Inner> testValue = 1;
        }

        message Inner {
            int32 innerValue = 1;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)

    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    MapType(
                        StringType(variable=True, bytes_=2_147_483_648),
                        StructType(
                            [
                                UnionType(
                                    [
                                        NullType(),
                                        IntType(bits=32, signed=True),
                                    ],
                                    name="innerValue",
                                    default=None,
                                )
                            ],
                            alias="_root.Inner",
                        ),
                    ),
                ],
                name="testValue",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_protobuf_converter_nested_message():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            message Inner {
                int32 innerValue = 1;
            }
            Inner outerValue = 2;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="_root.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="_root.Outer",
    )


def test_protobuf_converter_enum():
    protobuf_schema = """
        syntax = "proto3";
        message WeatherReport {
            Season season = 1;
        }
        enum Season {
            SPRING = 0;
            SUMMER = 1;
            AUTUMN = 2;
            WINTER = 3;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    EnumType(
                        ["SPRING", "SUMMER", "AUTUMN", "WINTER"],
                        alias="_root.Season",
                    ),
                ],
                name="season",
                default=None,
            )
        ],
        alias="_root.WeatherReport",
    )


def test_protobuf_converter_oneof():
    protobuf_schema = """
        syntax = "proto3";
        message Contact {
            oneof contact_type {
                string email = 1;
                int64 phone_number = 2;
            }
        }
    """

    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StringType(variable=True, bytes_=2_147_483_648, name="email"),
                    IntType(bits=64, signed=True, name="phone_number"),
                ],
                name="contact_type",
                default=None,
            )
        ],
        alias="_root.Contact",
    )


def test_protobuf_converter_doubly_nested_message():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            message Middle {
                message Inner {
                    int32 innerValue = 1;
                }
                Inner middleValue = 2;
            }
            Middle outerValue = 3;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)

    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    StructType(
                                        [
                                            UnionType(
                                                [
                                                    NullType(),
                                                    IntType(bits=32, signed=True),
                                                ],
                                                name="innerValue",
                                                default=None,
                                            )
                                        ],
                                        alias="_root.Inner",
                                    ),
                                ],
                                name="middleValue",
                                default=None,
                            )
                        ],
                        alias="_root.Middle",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="_root.Outer",
    )


def test_protobuf_converter_timestamp():
    protobuf_schema = """
    syntax = "proto3";
    import "google/protobuf/timestamp.proto";
    message Test {
        google.protobuf.Timestamp timestamp = 1;
    }
    """
    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    IntType(
                        bits=64,
                        signed=True,
                        logical="build.recap.Timestamp",
                        unit="nanosecond",
                        timezone="UTC",
                    ),
                ],
                name="timestamp",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_protobuf_converter_duration():
    protobuf_schema = """
    syntax = "proto3";
    import "google/protobuf/duration.proto";
    message Test {
        google.protobuf.Duration duration = 1;
    }
    """
    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    IntType(
                        bits=64,
                        signed=True,
                        logical="build.recap.Duration",
                        unit="nanosecond",
                    ),
                ],
                name="duration",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_protobuf_converter_nullvalue():
    protobuf_schema = """
    syntax = "proto3";
    import "google/protobuf/struct.proto";
    message Test {
        google.protobuf.NullValue null_value = 1;
    }
    """
    recap_schema = ProtobufConverter().to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            NullType(
                name="null_value",
                default=None,
            )
        ],
        alias="_root.Test",
    )


def test_protobuf_converter_self_reference():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            int32 value = 1;
            Outer next = 2;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)

    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    IntType(bits=32, signed=True),
                ],
                name="value",
                default=None,
            ),
            UnionType(
                [
                    NullType(),
                    ProxyType(
                        alias="_root.Outer",
                        registry=converter.registry,
                    ),
                ],
                name="next",
                default=None,
            ),
        ],
        alias="_root.Outer",
    )

    # Strip the alias since the resolved type won't have an alias.
    recap_schema.alias = None

    assert isinstance(recap_schema.fields[1], UnionType)
    assert isinstance(recap_schema.fields[1].types[1], ProxyType)
    assert recap_schema.fields[1].types[1].resolve() == recap_schema


def test_protobuf_converter_custom_namespace():
    protobuf_schema = """
        syntax = "proto3";
        message Outer {
            message Inner {
                int32 innerValue = 1;
            }
            Inner outerValue = 2;
        }
    """

    converter = ProtobufConverter("foo.bar")
    recap_schema = converter.to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="foo.bar.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="foo.bar.Outer",
    )


def test_protobuf_converter_package():
    protobuf_schema = """
        syntax = "proto3";
        package foo.bar.baz;
        message Outer {
            message Inner {
                int32 innerValue = 1;
            }
            Inner outerValue = 2;
        }
    """

    converter = ProtobufConverter()
    recap_schema = converter.to_recap(protobuf_schema)
    assert recap_schema == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="foo.bar.baz.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="foo.bar.baz.Outer",
    )


def test_protobuf_converter_separate_files():
    protobuf_schema_1 = """
        syntax = "proto3";
        package foo.blah;
        message Inner {
            int32 innerValue = 1;
        }
    """
    protobuf_schema_2 = """
        syntax = "proto3";
        message Outer {
            foo.blah.Inner outerValue = 1;
        }
    """

    converter = ProtobufConverter()

    # Ignore return result. Just calling to load Inner into registry.
    converter.to_recap(protobuf_schema_1)

    recap_schema_2 = converter.to_recap(protobuf_schema_2)

    assert recap_schema_2 == StructType(
        [
            UnionType(
                [
                    NullType(),
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    IntType(bits=32, signed=True),
                                ],
                                name="innerValue",
                                default=None,
                            )
                        ],
                        alias="foo.blah.Inner",
                    ),
                ],
                name="outerValue",
                default=None,
            )
        ],
        alias="_root.Outer",
    )



import json

from recap.types import (
    BoolType,
    FloatType,
    IntType,
    ListType,
    NullType,
    RecapType,
    StringType,
    StructType,
)


class JSONSchemaConverter:
    def to_recap(self, json_schema_str: str) -> StructType:
        json_schema = json.loads(json_schema_str)
        recap_schema = self._parse(json_schema)
        if not isinstance(recap_schema, StructType):
            raise ValueError("JSON schema must be an object")
        return recap_schema

    def _parse(self, json_schema: dict) -> RecapType:
        extra_attrs = {}
        if "description" in json_schema:
            extra_attrs["doc"] = json_schema["description"]
        if "default" in json_schema:
            extra_attrs["default"] = json_schema["default"]

        match json_schema:
            case {"type": "object", "properties": properties}:
                fields = []
                for name, prop in properties.items():
                    field = self._parse(prop)
                    if name not in json_schema.get("required", []):
                        field = field.make_nullable()
                    field.extra_attrs["name"] = name
                    fields.append(field)
                return StructType(fields, **extra_attrs)
            case {"type": "array", "items": items}:
                values = self._parse(items)
                return ListType(values, **extra_attrs)
            case {"type": "string", "format": "date"}:
                return StringType(
                    bytes_=9_223_372_036_854_775_807,
                    logical="org.iso.8601.Date",
                    **extra_attrs,
                )
            case {"type": "string", "format": "date-time"}:
                return StringType(
                    bytes_=9_223_372_036_854_775_807,
                    logical="org.iso.8601.DateTime",
                    **extra_attrs,
                )
            case {"type": "string", "format": "time"}:
                return StringType(
                    bytes_=9_223_372_036_854_775_807,
                    logical="org.iso.8601.Time",
                    **extra_attrs,
                )
            case {"type": "string"}:
                return StringType(bytes_=9_223_372_036_854_775_807, **extra_attrs)
            case {"type": "number"}:
                return FloatType(bits=64, **extra_attrs)
            case {"type": "integer"}:
                return IntType(bits=32, signed=True, **extra_attrs)
            case {"type": "boolean"}:
                return BoolType(**extra_attrs)
            case {"type": "null"}:
                return NullType(**extra_attrs)
            case _:
                raise ValueError(f"Unsupported JSON schema: {json_schema}")



from proto_schema_parser.ast import (
    Enum,
    EnumValue,
    Field,
    FieldCardinality,
    File,
    MapField,
    Message,
    OneOf,
    Package,
)
from proto_schema_parser.parser import Parser

from recap.types import (
    BoolType,
    BytesType,
    EnumType,
    FloatType,
    IntType,
    ListType,
    MapType,
    NullType,
    ProxyType,
    RecapType,
    RecapTypeRegistry,
    StringType,
    StructType,
    UnionType,
)

DEFAULT_NAMESPACE = "_root"
"""
Namespace to use when no namespace is specified in the schema.
"""


class ProtobufConverter:
    def __init__(self, namespace: str = DEFAULT_NAMESPACE) -> None:
        self.namespaces = [namespace]
        self.registry = RecapTypeRegistry()

    def to_recap(
        self,
        protobuf_schema_str: str,
    ) -> StructType:
        file = Parser().parse(protobuf_schema_str)
        root_message = self._parse(file)
        root_message = self._resolve_proxies(root_message)

        if not isinstance(root_message, StructType):
            raise ValueError("Protobuf schema must be a Message")

        return root_message

    def _resolve_proxies(
        self,
        recap_type: RecapType,
        resolved: set[str] | None = None,
        seen: set[str] | None = None,
    ) -> RecapType:
        resolved = resolved or set()
        seen = seen or set()

        # Detect circular references.
        if not isinstance(recap_type, ProxyType) and recap_type.alias is not None:
            seen.add(recap_type.alias)

        if (
            isinstance(recap_type, ProxyType)
            # Skip if already resolved.
            and recap_type.alias not in resolved
            # Skip if circular reference.
            and recap_type.alias not in seen
        ):
            extra_attrs = recap_type.extra_attrs
            alias = recap_type.alias
            recap_type = recap_type.resolve()
            recap_type.alias = alias
            recap_type.extra_attrs |= extra_attrs
            resolved.add(recap_type.alias)  # pyright: ignore[reportGeneralTypeIssues]

        if isinstance(recap_type, ListType):
            recap_type.values = self._resolve_proxies(recap_type.values, resolved, seen)
        elif isinstance(recap_type, MapType):
            recap_type.keys = self._resolve_proxies(recap_type.keys, resolved, seen)
            recap_type.values = self._resolve_proxies(recap_type.values, resolved, seen)
        elif isinstance(recap_type, UnionType):
            recap_type.types = [
                self._resolve_proxies(t, resolved, seen) for t in recap_type.types
            ]
        elif isinstance(recap_type, StructType):
            recap_type.fields = [
                self._resolve_proxies(f, resolved, seen) for f in recap_type.fields
            ]

        return recap_type

    def _parse(self, file: File) -> StructType:
        root_message = None

        # Find the namespace first, since we need it for message and enum aliases.
        packages = [p for p in file.file_elements if isinstance(p, Package)]

        if packages:
            self.namespaces.append(packages[0].name)

        # Now parse messages and enums.
        for file_element in file.file_elements:
            if isinstance(file_element, Message):
                struct_type = self._parse_message(file_element)
                root_message = root_message or struct_type
            elif isinstance(file_element, Enum):
                self._parse_enum(file_element)

        if root_message is None:
            raise ValueError("Protobuf schema must contain at least one Message")

        self.namespaces.pop()

        return root_message

    def _parse_message(self, message: Message) -> StructType:
        fields = []

        for element in message.elements:
            if isinstance(element, Field):
                recap_type = self._parse_field(element)
                fields.append(recap_type)
            elif isinstance(element, MapField):
                recap_type = self._parse_map_field(element)
                fields.append(recap_type)
            elif isinstance(element, OneOf):
                recap_type = self._parse_oneof(element)
                fields.append(recap_type)
            elif isinstance(element, Message):
                self._parse_message(element)
            elif isinstance(element, Enum):
                self._parse_enum(element)

        struct_type = StructType(
            fields=fields,
            alias=f"{self.namespaces[-1]}.{message.name}",
        )

        self.registry.register_alias(struct_type)

        return struct_type

    def _parse_field(self, field: Field) -> RecapType:
        recap_type = self._protobuf_type_to_recap_type(field.type)

        if field.cardinality == FieldCardinality.REPEATED:
            recap_type = ListType(values=recap_type)

        if field.cardinality != FieldCardinality.REQUIRED:
            if not isinstance(recap_type, NullType):
                recap_type = UnionType(types=[NullType(), recap_type], default=None)
            recap_type.extra_attrs["default"] = None

        recap_type.extra_attrs["name"] = field.name

        return recap_type

    def _parse_map_field(self, field: MapField) -> RecapType:
        key_type = self._protobuf_type_to_recap_type(field.key_type)

        # Proto map keys are always strings or integers.
        assert isinstance(key_type, IntType) or isinstance(key_type, StringType)

        value_type = self._protobuf_type_to_recap_type(field.value_type)
        map_type = UnionType(
            types=[
                NullType(),
                MapType(keys=key_type, values=value_type),
            ],
            default=None,
        )

        if field.name is not None:
            map_type.extra_attrs["name"] = field.name

        return map_type

    def _protobuf_type_to_recap_type(self, protobuf_type: str) -> RecapType:
        # NOTE: Protobuf doesn't support type aliases, so we don't need to
        # register aliases here (or in oneof, map, etc).
        match protobuf_type:
            case "string":
                return StringType(bytes_=2_147_483_648, variable=True)
            case "bool":
                return BoolType()
            case "bytes":
                return BytesType(bytes_=2_147_483_648, variable=True)
            case "int32":
                return IntType(bits=32)
            case "int64":
                return IntType(bits=64)
            case "uint32":
                return IntType(bits=32, signed=False)
            case "uint64":
                return IntType(bits=64, signed=False)
            case "sint32":
                return IntType(bits=32)
            case "sint64":
                return IntType(bits=64)
            case "float":
                return FloatType(bits=32)
            case "double":
                return FloatType(bits=64)
            case "fixed32":
                return IntType(bits=32, signed=False)
            case "fixed64":
                return IntType(bits=64, signed=False)
            case "sfixed32":
                return IntType(bits=32)
            case "sfixed64":
                return IntType(bits=64)
            # Handle some of Google's well-known types.
            # Technically, we should honor `import google/protobuf/*.proto` and
            # make sure the protos match what's expected, but assuming they match
            # is fine for now.
            case "google.protobuf.Timestamp":
                # https://protobuf.dev/reference/protobuf/google.protobuf/#timestamp
                # Note: Some protobuf values might overrun this since they support
                # from 0001-01-01T00:00:00Z to 9999-12-31T23:59:59.999999999Z.
                return IntType(
                    logical="build.recap.Timestamp",
                    bits=64,
                    signed=True,
                    unit="nanosecond",
                    timezone="UTC",
                )
            case "google.protobuf.Duration":
                # https://protobuf.dev/reference/protobuf/google.protobuf/#duration
                # Note: Some protobuf values might overrun this since they support
                # int64 seconds and int32 nanoseconds.
                return IntType(
                    bits=64,
                    logical="build.recap.Duration",
                    unit="nanosecond",
                )
            case "google.protobuf.NullValue":
                return NullType()
            case _:
                # Create a ProxyType and hope we have a type alias for this type.
                alias = (
                    protobuf_type
                    if "." in protobuf_type
                    else f"{self.namespaces[-1]}.{protobuf_type}"
                )
                return ProxyType(alias, self.registry)

    def _parse_oneof(self, oneof: OneOf) -> RecapType:
        types: list[RecapType] = [NullType()]
        for element in oneof.elements:
            if isinstance(element, Field):
                # !! HACK !!
                # Force oneof subfield to be required since we're unioning with null above.
                # Otherwise we end up with [NullType, [NullType, RecapType], ...].
                # This is safe to force because oneof subfields can't be repeated.
                original_cardinality = element.cardinality
                element.cardinality = FieldCardinality.REQUIRED
                recap_type = self._parse_field(element)
                element.cardinality = original_cardinality
                types.append(recap_type)
        return UnionType(types=types, name=oneof.name, default=None)

    def _parse_enum(self, enum: Enum) -> RecapType:
        symbols = []
        for element in enum.elements:
            if isinstance(element, EnumValue):
                symbols.append(element.name)
        enum_type = EnumType(
            symbols=symbols,
            alias=f"{self.namespaces[-1]}.{enum.name}",
        )
        self.registry.register_alias(enum_type)
        return enum_type



from recap.converters.json_schema import JSONSchemaConverter
from recap.types import (
    BoolType,
    FloatType,
    IntType,
    ListType,
    NullType,
    StringType,
    StructType,
    UnionType,
)


def test_all_basic_types():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "a_string": {"type": "string"},
            "a_number": {"type": "number"},
            "an_integer": {"type": "integer"},
            "a_boolean": {"type": "boolean"},
            "a_null": {"type": "null"}
        }
    }
    """
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
            name="a_string",
            default=None,
        ),
        UnionType(
            [NullType(), FloatType(bits=64)],
            name="a_number",
            default=None,
        ),
        UnionType(
            [NullType(), IntType(bits=32, signed=True)],
            name="an_integer",
            default=None,
        ),
        UnionType(
            [NullType(), BoolType()],
            name="a_boolean",
            default=None,
        ),
        UnionType(
            [NullType(), NullType()],
            name="a_null",
            default=None,
        ),
    ]


def test_nested_objects():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "an_object": {
                "type": "object",
                "properties": {
                    "a_string": {"type": "string"}
                }
            }
        }
    }
    """
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [
                NullType(),
                StructType(
                    [
                        UnionType(
                            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
                            name="a_string",
                            default=None,
                        ),
                    ],
                ),
            ],
            name="an_object",
            default=None,
        ),
    ]


def test_object_with_array_of_objects():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "an_array": {
                "type": "array",
                "items": {
                    "type": "object",
                    "properties": {
                        "a_string": {"type": "string"}
                    }
                }
            }
        }
    }
    """
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [
                NullType(),
                ListType(
                    StructType(
                        [
                            UnionType(
                                [
                                    NullType(),
                                    StringType(bytes_=9_223_372_036_854_775_807),
                                ],
                                name="a_string",
                                default=None,
                            )
                        ]
                    ),
                ),
            ],
            name="an_array",
            default=None,
        ),
    ]


def test_required_properties():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "required_property": {"type": "string"},
            "optional_property": {"type": "number"}
        },
        "required": ["required_property"]
    }
    """
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        StringType(bytes_=9_223_372_036_854_775_807, name="required_property"),
        UnionType(
            [NullType(), FloatType(bits=64)],
            name="optional_property",
            default=None,
        ),
    ]


def test_doc_attribute():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "a_string": {
                "type": "string",
                "description": "This is a string"
            }
        }
    }
    """
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
            name="a_string",
            doc="This is a string",
            default=None,
        ),
    ]


def test_name_attribute():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "a_string": {
                "type": "string",
                "title": "A String"
            }
        }
    }
    """
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [NullType(), StringType(bytes_=9_223_372_036_854_775_807)],
            name="a_string",
            default=None,
        ),
    ]


def test_default_attribute():
    json_schema = """
    {
        "type": "object",
        "properties": {
            "a_string": {
                "type": "string",
                "default": "Default value"
            }
        }
    }
    """
    struct_type = JSONSchemaConverter().to_recap(json_schema)
    assert isinstance(struct_type, StructType)
    assert struct_type.fields == [
        UnionType(
            [
                NullType(),
                StringType(bytes_=9_223_372_036_854_775_807),
            ],
            name="a_string",
            default="Default value",
        ),
    ]


def test_convert_date():
    converter = JSONSchemaConverter()
    schema = """
    {
        "type": "object",
        "properties": {
            "date": {
                "type": "string",
                "format": "date"
            }
        },
        "required": ["date"]
    }
    """
    result = converter.to_recap(schema)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], StringType)
    assert result.fields[0].logical == "org.iso.8601.Date"


def test_convert_datetime():
    converter = JSONSchemaConverter()
    schema = """
    {
        "type": "object",
        "properties": {
            "datetime": {
                "type": "string",
                "format": "date-time"
            }
        },
        "required": ["datetime"]
    }
    """
    result = converter.to_recap(schema)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], StringType)
    assert result.fields[0].logical == "org.iso.8601.DateTime"


def test_convert_time():
    converter = JSONSchemaConverter()
    schema = """
    {
        "type": "object",
        "properties": {
            "time": {
                "type": "string",
                "format": "time"
            }
        },
        "required": ["time"]
    }
    """
    result = converter.to_recap(schema)
    assert isinstance(result, StructType)
    assert isinstance(result.fields[0], StringType)
    assert result.fields[0].logical == "org.iso.8601.Time"
